/*---------------------------------------------------------------------------*\
  =========                 |
  \\      /  F ield         | OpenFOAM: The Open Source CFD Toolbox
   \\    /   O peration     |
    \\  /    A nd           | Copyright (C) 2011-2016 OpenFOAM Foundation
     \\/     M anipulation  | Copyright (C) 2017 OpenCFD Ltd.
-------------------------------------------------------------------------------
License
    This file is part of OpenFOAM.

    OpenFOAM is free software: you can redistribute it and/or modify it
    under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    OpenFOAM is distributed in the hope that it will be useful, but WITHOUT
    ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
    FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
    for more details.

    You should have received a copy of the GNU General Public License
    along with OpenFOAM.  If not, see <http://www.gnu.org/licenses/>.

\*---------------------------------------------------------------------------*/

#include "vtkPVFoam.H"

// OpenFOAM includes
#include "Cloud.H"
#include "IOobjectList.H"
#include "vtkPVFoamReader.h"

// VTK includes
#include "vtkDataArraySelection.h"
#include "vtkPolyData.h"
#include "vtkUnstructuredGrid.h"

// Templates (only needed here)
#include "vtkPVFoamFieldTemplates.C"

// * * * * * * * * * * * * * * * Member Functions  * * * * * * * * * * * * * //

void Foam::vtkPVFoam::pruneObjectList
(
    IOobjectList& objects,
    const hashedWordList& retain
)
{
    if (retain.empty())
    {
        objects.clear();
    }

    // only retain specified fields
    forAllIter(IOobjectList, objects, iter)
    {
        if (!retain.found(iter()->name()))
        {
            objects.erase(iter);
        }
    }
}


void Foam::vtkPVFoam::convertVolFields
(
    vtkMultiBlockDataSet* output
)
{
    const fvMesh& mesh = *meshPtr_;

    const bool interpFields = reader_->GetInterpolateVolFields();
    hashedWordList selectedFields = getSelected
    (
        reader_->GetVolFieldSelection()
    );

    if (selectedFields.empty())
    {
        return;
    }

    // Get objects (fields) for this time - only keep selected fields
    // the region name is already in the mesh db
    IOobjectList objects(mesh, dbPtr_().timeName());
    pruneObjectList(objects, selectedFields);

    if (objects.empty())
    {
        return;
    }

    if (debug)
    {
        Info<< "<beg> convert volume fields" << endl;
        forAllConstIter(IOobjectList, objects, iter)
        {
            Info<< "  " << iter()->name()
                << " == " << iter()->objectPath() << endl;
        }
        printMemory();
    }


    PtrList<patchInterpolator> interpLst;

    if (interpFields)
    {
        interpLst.setSize(mesh.boundaryMesh().size());

        forAll(interpLst, i)
        {
            interpLst.set
            (
                i,
                new PrimitivePatchInterpolation<primitivePatch>
                (
                    mesh.boundaryMesh()[i]
                )
            );
        }
    }

    convertVolFields<scalar>(mesh, interpLst, objects, output);
    convertVolFields<vector>(mesh, interpLst, objects, output);
    convertVolFields<sphericalTensor>(mesh, interpLst, objects, output);
    convertVolFields<symmTensor>(mesh, interpLst, objects, output);
    convertVolFields<tensor>(mesh, interpLst, objects, output);

    convertDimFields<scalar>(mesh, interpLst, objects, output);
    convertDimFields<vector>(mesh, interpLst, objects, output);
    convertDimFields<sphericalTensor>(mesh, interpLst, objects, output);
    convertDimFields<symmTensor>(mesh, interpLst, objects, output);
    convertDimFields<tensor>(mesh, interpLst, objects, output);

    if (debug)
    {
        Info<< "<end> convert volume fields" << endl;
        printMemory();
    }
}


void Foam::vtkPVFoam::convertPointFields
(
    vtkMultiBlockDataSet* output
)
{
    const fvMesh& mesh = *meshPtr_;

    hashedWordList selectedFields = getSelected
    (
        reader_->GetPointFieldSelection()
    );

    if (selectedFields.empty())
    {
        if (debug)
        {
            Info<< "no point fields selected" << endl;
        }
        return;
    }

    // Get objects (fields) for this time - only keep selected fields
    // the region name is already in the mesh db
    IOobjectList objects(mesh, dbPtr_().timeName());
    pruneObjectList(objects, selectedFields);

    if (objects.empty())
    {
        return;
    }

    if (debug)
    {
        Info<< "<beg> convert volume -> point fields" << endl;
        forAllConstIter(IOobjectList, objects, iter)
        {
            Info<< "  " << iter()->name()
                << " == " << iter()->objectPath() << endl;
        }
        printMemory();
    }

    // Construct interpolation on the raw mesh
    const pointMesh& pMesh = pointMesh::New(mesh);

    convertPointFields<scalar>(pMesh, objects, output);
    convertPointFields<vector>(pMesh, objects, output);
    convertPointFields<sphericalTensor>(pMesh, objects, output);
    convertPointFields<symmTensor>(pMesh, objects, output);
    convertPointFields<tensor>(pMesh, objects, output);

    if (debug)
    {
        Info<< "<end> convert volume -> point fields" << endl;
        printMemory();
    }
}


void Foam::vtkPVFoam::convertLagrangianFields
(
    vtkMultiBlockDataSet* output
)
{
    arrayRange& range = arrayRangeLagrangian_;
    const fvMesh& mesh = *meshPtr_;

    hashedWordList selectedFields = getSelected
    (
        reader_->GetLagrangianFieldSelection()
    );

    if (selectedFields.empty())
    {
        return;
    }

    if (debug)
    {
        Info<< "<beg> convert Lagrangian fields" << endl;
        printMemory();
    }

    for (int partId = range.start(); partId < range.end(); ++partId)
    {
        const word  cloudName = getPartName(partId);
        const label datasetNo = partDataset_[partId];

        if (!partStatus_[partId] || datasetNo < 0)
        {
            continue;
        }

        // Get the Lagrangian fields for this time and this cloud
        // but only keep selected fields
        // the region name is already in the mesh db
        IOobjectList objects
        (
            mesh,
            dbPtr_().timeName(),
            cloud::prefix/cloudName
        );
        pruneObjectList(objects, selectedFields);

        if (objects.empty())
        {
            continue;
        }

        if (debug)
        {
            Info<< "converting OpenFOAM lagrangian fields" << endl;
            forAllConstIter(IOobjectList, objects, iter)
            {
                Info<< "  " << iter()->name()
                    << " == " << iter()->objectPath() << endl;
            }
        }

        convertLagrangianFields<label>(objects, output, datasetNo);
        convertLagrangianFields<scalar>(objects, output, datasetNo);
        convertLagrangianFields<vector>(objects, output, datasetNo);
        convertLagrangianFields<sphericalTensor>(objects, output, datasetNo);
        convertLagrangianFields<symmTensor>(objects, output, datasetNo);
        convertLagrangianFields<tensor>(objects, output, datasetNo);
    }

    if (debug)
    {
        Info<< "<end> convert Lagrangian fields" << endl;
        printMemory();
    }
}


// ************************************************************************* //
